/* start.S -- standalone entry point and auxv processor.
   Copyright 2007 BitWagon Software LLC.  All rights reserved.
   Licensed under GNU General Public License, version 2 [GPLv2].
*/
#include <asm/unistd.h>

#define AT_NULL   0
#define AT_IGNORE 1
#define AT_PHDR   3
#define AT_PHENT  4
#define AT_PHNUM  5
#define AT_PAGESZ 6
#define AT_BASE   7
#define AT_FLAGS  8
#define AT_ENTRY  9

slots= (1<<AT_PHDR) | (1<<AT_PHENT) | (1<<AT_PHNUM) | (1<<AT_PAGESZ)
slots= slots | (1<<AT_BASE) | (1<<AT_FLAGS) | (1<<AT_ENTRY)

sz_auxv= 8

_start: .globl _start
        subl %ecx,%ecx  // 0
	movl $slots,%edx
        movl %esp,%esi
        movl %esp,%edi
        call do_auxv  // clear bits in edx according to existing auxv slots
        testl %edx,%edx; je L60  // slots already present

        movl %esp,%esi
L50:
        shrl %edx  // Carry = bottom bit
        sbbl %eax,%eax  // -1 or 0
        subl %eax,%ecx  // count of 1 bits that remained in edx
        leal (%esp,%eax,sz_auxv),%esp  // allocate one auxv slot, if needed
	testl %edx,%edx; jne L50

        movl %esp,%edi
        call do_auxv  // move; fill new auxv slots with AT_IGNORE
L60:
	movl %esp,%eax
	call my_main
	hlt

// called twice:
//  1st with esi==edi, ecx=0, edx= bitmap of slots needed: just update edx.
//  2nd with esi!=edi, ecx= slot_count: move, then append AT_IGNORE slots
// entry: esi= src = &argc; edi= dst; ecx= # slots wanted; edx= bits wanted
// exit:  edi= &auxtab; edx= bits still needed; ecx= 0; ebx= &envp[0]; esi= &auxv[N]
do_auxv:
        // cld

L10:  // move argc+argv
        lodsl
        stosl
        testl %eax,%eax
        jne L10

L20:  // move envp
        lodsl
        stosl
        testl %eax,%eax
        jne L20

        pushl %edi  // return value
L30:  // process auxv
        lodsl  // a_type
        stosl
        cmpl $32,%eax; jae L32  // prevent aliasing by 'btr' when 32<=a_type
        btr %eax,%edx  // no longer need a slot of type eax  [Carry only]
L32:
        testl %eax,%eax  // AT_NULL ?
        lodsl
        stosl
        jnz L30  // a_type != AT_NULL

        subl $8,%edi  // backup to AT_NULL
        shll %ecx  // two words per auxv
        incl %eax  // convert 0 to AT_IGNORE
        rep; stosl  // allocate and fill
        decl %eax  // convert AT_IGNORE to AT_NULL
        stosl  // re-terminate with AT_NULL
        stosl
        movl %edi,%esi  // &auxv[N]

        popl %edi  // &auxtab
        ret

jump_exit: .globl jump_exit #(*)(void **esp, int, void *target) __attribute__((regparm(3)))
	movl %eax,%esp
	jmp *%ecx

	.section .note.GNU-stack,"",@progbits
