/* syscall.S -- standalone implementation of system calls.
   Copyright 2007 BitWagon Software LLC.  All rights reserved.
   Licensed under GNU General Public License, version 2 [GPLv2].
*/
#include <asm/unistd.h>

AT_PLATFORM= 15  # value is &char[]
MAX_ARG_PAGES= 32
PAGE_SHIFT= 12
PAGE_MASK = ~0<<PAGE_SHIFT  # AND clears the offset-within-page bits

spin: .globl spin
0:
	pause
	jmp 0b
	ret
	
tail_merge: .globl tail_merge  # (*)(&argc, target, c_end)  __attribute__((regparm(3)))
	movl %ecx,%ebx  # original c_end
	orl $0xfff,%ecx  # &last_byte_on_page
	leal 1-(MAX_ARG_PAGES<<PAGE_SHIFT)(%ecx),%esp
	movl %esp,a_argc

	movl %eax,%esi  # src
	movl %esp,%edi  # dst
	pushl %edx  # target address
	cld
	movsl # argc

	movl %edi,%ecx
	subl %esi,%ecx  # relocation amount
	call reloc_vecs_i
	movl %ebx,%ecx
	subl %esi,%ecx
	rep; movsb

	lea 4(%esp),%eax  # new &argc
	ret  # goto target

reloc_vecs: .globl reloc_vecs  # (*)(ptr, delta) __attribute__((regparm(2)))
	push %esi; push %edi
	movl %eax,%esi  // src
	movl %eax,%edi  // dst same as src
	movl %edx,%ecx  // delta
	call reloc_vecs_i
	pop %edi; pop %esi
	ret

reloc_vecs_i:
	call reloc_strvec  # argv
	call reloc_strvec  # envp
reloc_auxv:
	lodsl; movl %eax,%edx
	stosl; cmpl $AT_PLATFORM,%eax
	lodsl; jne L20; and $3,%eax; lea 4+8(%edi,%eax),%eax  # FIXME 2007-08-23
L20:
	stosl; testl %edx,%edx; jne reloc_auxv
	ret

reloc_strvec:
	lodsl; testl %eax,%eax; je L10; addl %ecx,%eax
	stosl; jmp reloc_strvec
L10:
	stosl
	ret


isync: .globl isync  # isync(void *, unsigned)  /* args unused for now */
	pushl %ebx; subl %eax,%eax; cpuid
	popl %ebx
	ret

my_memset: .globl my_memset
	movl 1*4(%esp),%edx  # ptr
	movl 2*4(%esp),%eax  # val
	movl 3*4(%esp),%ecx  # len
	push %edi
	movl %edx,%edi
	cld; rep; stosb
	pop %edi
	movl %edx,%eax  # result= original ptr
	ret

0:
	ret
my_mprotect:.globl my_mprotect; .type my_mprotect,@function  # minimum page cover of the interval
	movl 2*4(%esp),%ecx  # length
	movl 1*4(%esp),%eax  # addr
	testl %ecx,%ecx; je 0b
	movl $ PAGE_MASK,%edx
	leal -1(%eax,%ecx),%ecx  # &last affected
	andl %edx,%eax  # round down
	andl %edx,%ecx
	subl %eax,%ecx  # page-crossing length
	subl %edx,%ecx  # total length
	movl %eax,1*4(%esp)  # new addr
	movl %ecx,2*4(%esp)  # new length
mprotect: .globl mprotect
	movb $__NR_mprotect,%al; jmp sys3

	.balign 4

chdir: .globl chdir
	movb $__NR_chdir,%al; jmp sys3

close: .globl close
	movb $__NR_close,%al; jmp sys3

dup: .globl dup
	movb $__NR_dup,%al; jmp sys3

dup2: .globl dup2
	movb $__NR_dup2,%al; jmp sys3

_exit: .globl _exit
	movb $__NR_exit,%al; jmp sys3

fchdir: .globl fchdir
	movb $__NR_fchdir,%al; jmp sys3

fcntl: .globl fcntl
	movb $__NR_fcntl,%al; jmp sys3

fork: .globl fork
	movb $__NR_fork,%al; jmp sys3

getpid: .globl getpid
	movb $__NR_getpid,%al; jmp sys3

gettimeofday: .globl gettimeofday
	movb $__NR_gettimeofday,%al; jmp sys3

kill: .globl kill
	movb $__NR_kill, %al; jmp sys3

lseek: .globl lseek
	movb $__NR_lseek,%al; jmp sys3

my_lstat64: .globl my_lstat64
	movb $__NR_lstat64,%al; jmp sys3

mkdir: .globl mkdir
	movb $__NR_mkdir,%al; jmp sys3

my_mknod: .globl my_mknod
mknod: .globl mknod
	movb $__NR_mknod,%al; jmp sys3

munmap: .globl munmap
	movb $__NR_munmap,%al; jmp sys3

nanosleep: .globl nanosleep
	movb $__NR_nanosleep,%al; jmp sys3

open: .globl open
	movb $__NR_open,%al; jmp sys3

pause: .globl pause
	movb $__NR_pause,%al; jmp sys3

pipe: .globl pipe
	movb $__NR_pipe,%al; jmp sys3

read: .globl read
	movb $__NR_read,%al; jmp sys3

readlink: .globl readlink
	movb $__NR_readlink,%al; jmp sys3

my_stat64: .globl my_stat64
	movb $__NR_stat64,%al; jmp sys3

waitpid: .globl waitpid
	movb $__NR_waitpid,%al; jmp sys3

write: .globl write
	movb $__NR_write,%al; //jmp sys3

sys3:
	push %ebp; movl %esp,%ebp  # frame for debugging
	push %ebx;
	movl 2*4(%ebp),%ebx
	movl 3*4(%ebp),%ecx
	movl 4*4(%ebp),%edx
sys3go:
	movzbl %al,%eax
	int $0x80
	pop %ebx
	pop %ebp
	ret

mmap: .globl mmap
	push %ebp; movl %esp,%ebp  # frame for debugging
	pushl %ebx
	leal 2*4(%ebp),%ebx
	movb $__NR_mmap,%al; jmp sys3go

rt_sigaction: .globl rt_sigaction; .type rt_sigaction,@function; .size rt_sigaction,4  # sig, newact, oldact, sigsiz
	movb $ __NR_rt_sigaction,%al; jmp sys5

lseek64: .globl lseek64  # int (*)(int fd, uint hi, uint lo, loff_t *result, uint whence)
	movb $__NR__llseek,%al; jmp sys5

ptrace: .globl ptrace
	movb $__NR_ptrace,%al
sys5:
	push %ebp; mov %esp,%ebp  # frame for debugging
	push %ebx; push %esi; push %edi
	movl 2*4(%ebp),%ebx
	movl 3*4(%ebp),%ecx
	movl 4*4(%ebp),%edx
	movl 5*4(%ebp),%esi
	movl 6*4(%ebp),%edi
sys5go:
	movzbl %al,%eax
	int $0x80
	pop %edi; pop %esi; pop %ebx
	pop %ebp
	ret

	.section .note.GNU-stack,"",@progbits
